from optimum.onnxruntime import ORTModelForFeatureExtraction
from transformers import AutoTokenizer
from ezlocalai.Helpers import get_tokens, chunk_content
from huggingface_hub import hf_hub_download

import torch
import os


class Embedding:
    def __init__(self):
        model_dir = os.path.join(
            os.getcwd(),
            "models",
            "models--hooman650--bge-m3-onnx-o4",
            "snapshots",
            "848e8bc2408aad2c8849c7be9475f7aec3ee781a",
        )
        if not os.path.exists(model_dir):
            os.makedirs(model_dir, exist_ok=True)
        # if model_optimized.onnx.data does not exist, download the model.
        files = [
            "config.json",
            "model_optimized.onnx",
            "model_optimized.onnx.data",
            "ort_config.json",
            "sentencepiece.bpe.model",
            "special_tokens_map.json",
            "tokenizer.json",
            "tokenizer_config.json",
        ]
        for file in files:
            if not os.path.exists(os.path.join(model_dir, file)):
                hf_hub_download(
                    repo_id="hooman650/bge-m3-onnx-o4",
                    filename=file,
                    local_dir=model_dir,
                )
        self.model = ORTModelForFeatureExtraction.from_pretrained(
            "hooman650/bge-m3-onnx-o4",
            local_dir=model_dir,
            cache_dir=os.path.join(os.getcwd(), "models"),
        )
        self.tokenizer = AutoTokenizer.from_pretrained(
            "hooman650/bge-m3-onnx-o4",
            local_dir=model_dir,
            cache_dir=os.path.join(os.getcwd(), "models"),
        )

    def get_embeddings(self, input):
        tokens = get_tokens(input)
        sentences = chunk_content(input)
        encoded_input = self.tokenizer(
            sentences, padding=True, truncation=True, return_tensors="pt"
        ).to("cuda" if torch.cuda.is_available() else "cpu")
        out = self.model(**encoded_input, return_dict=True).last_hidden_state
        dense_vecs = torch.nn.functional.normalize(out[:, 0], dim=-1)
        embeddings = dense_vecs.cpu().detach().numpy().tolist()
        data = []
        for i, embedding in enumerate(embeddings):
            data.append(
                {
                    "object": "embedding",
                    "index": i,
                    "embedding": embedding,
                }
            )
        return {
            "object": "list",
            "data": data,
            "model": "bge-m3",
            "usage": {"prompt_tokens": tokens, "total_tokens": tokens},
        }



try:
    import spacy
except:
    import sys
    import subprocess

    subprocess.check_call([sys.executable, "-m", "pip", "install", "spacy"])
    import spacy
from typing import List
import tiktoken


def get_tokens(text: str) -> int:
    encoding = tiktoken.get_encoding("cl100k_base")
    num_tokens = len(encoding.encode(text))
    return num_tokens


def chunk_content(text: str) -> List[str]:
    try:
        sp = spacy.load("en_core_web_sm")
    except:
        spacy.cli.download("en_core_web_sm")
        sp = spacy.load("en_core_web_sm")
    sp.max_length = 99999999999999999999999
    doc = sp(text)
    sentences = list(doc.sents)
    content_chunks = [str(sentence).strip() for sentence in sentences]
    return content_chunks



import streamlit as st
import openai
import requests
import base64
import os
import re
from datetime import datetime
from dotenv import load_dotenv
import time

load_dotenv()
# Show logo but resize to the screen https://devxt.com/wp-content/uploads/2023/01/Logo-1024x316.png
st.image(
    "https://devxt.com/wp-content/uploads/2023/01/Logo-1024x316.png",
    width=300,
)

EZLOCALAI_SERVER = os.getenv("EZLOCALAI_URL", "http://localhost:8091")
EZLOCALAI_API_KEY = os.getenv("EZLOCALAI_API_KEY", "none")
DEFAULT_LLM = os.getenv("DEFAULT_MODEL", "TheBloke/phi-2-dpo-GGUF")
VISION_MODEL = os.getenv("VISION_MODEL", None)
SD_MODEL = os.getenv("SD_MODEL", "stabilityai/sdxl-turbo")
IMG_ENABLED = os.getenv("IMG_ENABLED", "false")
WHISPER_MODEL = os.getenv("WHISPER_MODEL", "base.en")

if "/" in DEFAULT_LLM:
    link_to_model = f"https://huggingface.co/{DEFAULT_LLM}"
else:
    link_to_model = f"https://huggingface.co/models?search={DEFAULT_LLM}"
st.markdown(
    f"""
    [![GitHub](https://img.shields.io/badge/GitHub-ezLocalai-blue?logo=github&style=plastic)](https://github.com/DevXT-LLC/ezlocalai) [![Dockerhub](https://img.shields.io/badge/Docker-ezlocalai-blue?logo=docker&style=plastic)](https://hub.docker.com/r/joshxt/ezlocalai)
    ## ezLocal.ai Demo
    **Language model:** [{DEFAULT_LLM}]({link_to_model})
    """
)
if VISION_MODEL:
    st.markdown(
        f"""
        **Vision model:** [{VISION_MODEL}](https://huggingface.co/{VISION_MODEL})
        """
    )
if IMG_ENABLED.lower() == "true":
    st.markdown(
        f"""
        **Image Generation model:** [{SD_MODEL}](https://huggingface.co/{SD_MODEL})
        """
    )
openai.base_url = f"{EZLOCALAI_SERVER}/v1/"
openai.api_key = EZLOCALAI_API_KEY if EZLOCALAI_API_KEY else EZLOCALAI_SERVER
HEADERS = {
    "Content-Type": "application/json",
    "Authorization": f"{EZLOCALAI_API_KEY}",
    "ngrok-skip-browser-warning": "true",
}
waiting_for_server = False


def get_voices():
    return ["Morgan_Freeman", "DukeNukem", "HAL9000"]
    # Commented for speed, but this works if we want to get the voices from the server
    """
    global EZLOCALAI_SERVER
    global HEADERS
    global waiting_for_server
    while True:
        try:
            voices = requests.get(
                f"{EZLOCALAI_SERVER}/v1/audio/voices", headers=HEADERS
            )
            voices = voices.json()["voices"]
            waiting_for_server = False
            return voices
        except:
            if waiting_for_server == False:
                st.spinner("Waiting for server to start...")
            waiting_for_server = True
            time.sleep(1)
    """


def display_content(content):
    global EZLOCALAI_SERVER
    global HEADERS
    outputs_url = f"{EZLOCALAI_SERVER}/outputs/"
    os.makedirs("outputs", exist_ok=True)
    if "http://localhost:8091/outputs/" in content:
        if outputs_url != "http://localhost:8091/outputs/":
            content = content.replace("http://localhost:8091/outputs/", outputs_url)
    if "<audio controls>" in content or " " not in content:
        try:
            audio_response = content.split("data:audio/wav;base64,")[1].split('" type')[
                0
            ]
        except:
            audio_response = content
        file_name = f"outputs/{datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}.wav"
        with open(file_name, "wb") as fh:
            fh.write(base64.b64decode(audio_response))
        st.audio(file_name, format="audio/wav", start_time=0)
    if outputs_url in content:
        urls = re.findall(f"{re.escape(outputs_url)}[^\"' ]+", content)
        urls = urls[0].split("\n\n")
        for url in urls:
            file_name = url.split("/")[-1]
            url = f"{outputs_url}{file_name}"
            data = requests.get(url, headers=HEADERS).content
            if url.endswith(".jpg") or url.endswith(".png"):
                content = content.replace(url, "")
                st.image(data, use_column_width=True)
            elif url.endswith(".mp4"):
                content = content.replace(url, "")
                st.audio(data, format="audio/mp4", start_time=0)
            elif url.endswith(".wav"):
                content = content.replace(url, "")
                st.audio(data, format="audio/wav", start_time=0)
    st.markdown(content, unsafe_allow_html=True)


show_advanced_options = st.checkbox(
    "Show Advanced Options", key="show_advanced_options"
)
if show_advanced_options:
    SYSTEM_MESSAGE = st.text_area(
        "System Prompt",
        "",
    )
    DEFAULT_MAX_TOKENS = st.number_input(
        "Max Output Tokens", min_value=10, max_value=300000, value=1024
    )
    DEFAULT_TEMPERATURE = st.number_input(
        "Temperature", min_value=0.0, max_value=1.0, value=0.5
    )
    DEFAULT_TOP_P = st.number_input("Top P", min_value=0.0, max_value=1.0, value=0.9)
else:
    SYSTEM_MESSAGE = ""
    DEFAULT_MAX_TOKENS = 1024
    DEFAULT_TEMPERATURE = 0.5
    DEFAULT_TOP_P = 0.9
with st.form("chat"):
    voice_drop_down = st.selectbox(
        "Text-to-Speech Response Voice", ["None"] + get_voices(), index=0
    )
    uploaded_file = st.file_uploader("Upload an image")
    prompt = st.text_area("Your Message:", "Describe each stage of this image.")
    send = st.form_submit_button("Send")
    if prompt != "" and send:
        start_time = time.time()
        st.markdown("---")
        st.spinner("Thinking...")
        messages = []
        if SYSTEM_MESSAGE != "":
            messages.append({"role": "system", "content": SYSTEM_MESSAGE})
        if uploaded_file:
            messages.append(
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": prompt},
                        {
                            "type": f"{uploaded_file.type.split('/')[0]}_url",
                            f"{uploaded_file.type.split('/')[0]}_url": {
                                "url": f"data:{uploaded_file.type};base64,{base64.b64encode(uploaded_file.read()).decode('utf-8')}",
                            },
                        },
                    ],
                },
            )
            if uploaded_file.type.startswith("image"):
                st.image(uploaded_file, use_column_width=True)
        if messages == []:
            messages = [
                {"role": "user", "content": prompt},
            ]
        extra_body = None if voice_drop_down == "None" else {"voice": voice_drop_down}
        response = openai.chat.completions.create(
            model=DEFAULT_LLM,
            messages=messages,
            temperature=DEFAULT_TEMPERATURE,
            max_tokens=DEFAULT_MAX_TOKENS,
            top_p=DEFAULT_TOP_P,
            stream=False,
            extra_body=extra_body,
        )
        display_content(response.choices[0].message.content)
        end_time = time.time()  # Record the end time
        elapsed_time = end_time - start_time
        # If response time is longer than 60 seconds, split the response time into minutes and seconds
        if elapsed_time > 60:
            minutes = int(elapsed_time // 60)
            seconds = elapsed_time % 60
            st.success(f"Response time: {minutes} minutes and {seconds:.2f} seconds")
        else:
            st.success(f"Response time: {elapsed_time:.2f} seconds")
        st.balloons()



try:
    from deepseek_vl.models import VLChatProcessor
except:
    VLChatProcessor = None
from transformers import AutoModelForCausalLM
from datetime import datetime
import requests
import torch
import PIL.Image
import uuid
import os
import base64
from ezlocalai.Helpers import get_tokens


class VLM:
    def __init__(self, model="deepseek-ai/deepseek-vl-1.3b-chat"):
        self.model = model.split("/")[-1]
        self.params = {}
        os.makedirs(os.path.join(os.getcwd(), "outputs"), exist_ok=True)
        try:
            self.vl_chat_processor = VLChatProcessor.from_pretrained(model)
            self.tokenizer = self.vl_chat_processor.tokenizer
            self.vl_gpt = AutoModelForCausalLM.from_pretrained(
                model,
                trust_remote_code=True,
                cache_dir=os.path.join(os.getcwd(), "models"),
            )
            self.vl_gpt = self.vl_gpt.to(torch.bfloat16).cuda().eval()
        except:
            self.vl_chat_processor = None
            self.tokenizer = None
            self.vl_gpt = None

    def chat(self, messages, **kwargs):
        pil_images = []
        images = []
        prompt = ""
        for message in messages:
            if isinstance(message["content"], str):
                role = message["role"] if "role" in message else "User"
                if role.lower() == "user":
                    prompt += f"{message['content']}\n\n"
                if role.lower() == "system":
                    prompt = f"System: {message['content']}\n\nUser: {prompt}"
            if isinstance(message["content"], list):
                for msg in message["content"]:
                    if "text" in msg:
                        role = message["role"] if "role" in message else "User"
                        if role.lower() == "user":
                            prompt += f"{msg['text']}\n\n"
                    if "image_url" in msg:
                        url = (
                            msg["image_url"]["url"]
                            if "url" in msg["image_url"]
                            else msg["image_url"]
                        )
                        image_path = f"./outputs/{uuid.uuid4().hex}.jpg"
                        if url.startswith("http"):
                            image = requests.get(url).content
                        else:
                            file_type = url.split(",")[0].split("/")[1].split(";")[0]
                            if file_type == "jpeg":
                                file_type = "jpg"
                            image_path = f"./outputs/{uuid.uuid4().hex}.{file_type}"
                            image = base64.b64decode(url.split(",")[1])
                        with open(image_path, "wb") as f:
                            f.write(image)
                        images.append(image_path)
                        pil_img = PIL.Image.open(image_path)
                        pil_img = pil_img.convert("RGB")
                        pil_images.append(pil_img)
        if len(images) > 0:
            for image in images:
                prompt = f"<image_placeholder> {prompt}"
        conversation = [
            {"role": "User", "content": prompt, "images": images},
            {"role": "Assistant", "content": ""},
        ]
        prepare_inputs = self.vl_chat_processor(
            conversations=conversation, images=pil_images, force_batchify=True
        ).to(self.vl_gpt.device)
        inputs_embeds = self.vl_gpt.prepare_inputs_embeds(**prepare_inputs)
        outputs = self.vl_gpt.language_model.generate(
            inputs_embeds=inputs_embeds,
            attention_mask=prepare_inputs.attention_mask,
            pad_token_id=self.tokenizer.eos_token_id,
            bos_token_id=self.tokenizer.bos_token_id,
            eos_token_id=self.tokenizer.eos_token_id,
            max_new_tokens=(
                1024 if "max_tokens" not in kwargs else int(kwargs["max_tokens"])
            ),
            do_sample=False,
            use_cache=True,
        )
        answer = self.tokenizer.decode(
            outputs[0].cpu().tolist(), skip_special_tokens=True
        )
        completion_tokens = get_tokens(answer)
        prompt_tokens = get_tokens(
            " ".join([message["content"] for message in conversation])
        )
        total_tokens = completion_tokens + prompt_tokens
        data = {
            "choices": [
                {
                    "finish_reason": "stop",
                    "index": 0,
                    "message": {"content": answer, "role": "assistant"},
                    "logprobs": None,
                }
            ],
            "created": datetime.now().isoformat(),
            "id": f"chatcmpl-{uuid.uuid4().hex}",
            "model": self.model,
            "object": "chat.completion",
            "usage": {
                "completion_tokens": completion_tokens,
                "prompt_tokens": prompt_tokens,
                "total_tokens": total_tokens,
            },
        }
        return data

    def completion(self, prompt, **kwargs):
        messages = [
            {"role": "User", "content": prompt},
        ]
        completion = self.chat(
            messages=messages,
            max_tokens=kwargs["max_tokens"] if "max_tokens" in kwargs else 1024,
        )
        data = {
            "choices": [
                {
                    "finish_reason": "length",
                    "index": 0,
                    "logprobs": None,
                    "text": completion["choices"][0]["message"]["content"],
                }
            ],
            "created": datetime.now().isoformat(),
            "id": f"cmpl-{uuid.uuid4().hex}",
            "model": self.model,
            "object": "text_completion",
            "usage": {
                "completion_tokens": completion["usage"]["completion_tokens"],
                "prompt_tokens": completion["usage"]["prompt_tokens"],
                "total_tokens": completion["usage"]["total_tokens"],
            },
        }
        return data

    def describe_image(self, image_url):
        messages = [
            {
                "role": "User",
                "content": [
                    {"type": "image_url", "image_url": {"url": image_url}},
                    {
                        "type": "text",
                        "text": "Describe each stage of this image.",
                    },
                ],
            },
        ]
        response = self.chat(
            messages=messages,
        )
        return response["choices"][0]["message"]["content"]

    def models(self):
        return [
            "deepseek-ai/deepseek-vl-1.3b-chat",
            "deepseek-ai/deepseek-vl-7b-chat",
        ]



from fastapi import (
    FastAPI,
    Depends,
    HTTPException,
    Header,
    Request,
    Form,
    UploadFile,
    File,
)
from fastapi.responses import StreamingResponse
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from pydantic import BaseModel
from typing import List, Dict, Union, Optional
from Pipes import Pipes
import base64
import os
import logging
import uuid
import time
from dotenv import load_dotenv

load_dotenv()
DEFAULT_MODEL = os.getenv("DEFAULT_MODEL", "TheBloke/phi-2-dpo-GGUF")
WHISPER_MODEL = os.getenv("WHISPER_MODEL", "base")
logging.basicConfig(
    level=os.environ.get("LOGLEVEL", "INFO"),
    format="%(asctime)s | %(levelname)s | %(message)s",
)
pipe = Pipes()
logging.info(f"[ezlocalai] Server is ready.")

app = FastAPI(title="ezlocalai Server", docs_url="/")
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
NGROK_TOKEN = os.environ.get("NGROK_TOKEN", "")
if NGROK_TOKEN:
    from pyngrok import ngrok

    ngrok.set_auth_token(NGROK_TOKEN)
    public_url = ngrok.connect(8091)
    logging.info(f"[ngrok] Public Tunnel: {public_url.public_url}")
    ngrok_url = public_url.public_url

    def get_ngrok_url():
        global ngrok_url
        return ngrok_url

else:

    def get_ngrok_url():
        return "http://localhost:8091"


app.mount("/outputs", StaticFiles(directory="outputs"), name="outputs")


def verify_api_key(authorization: str = Header(None)):
    encryption_key = os.environ.get("EZLOCALAI_API_KEY", "")
    if encryption_key:
        if authorization is None:
            raise HTTPException(
                status_code=401, detail="Authorization header is missing"
            )
        try:
            if "bearer " in authorization.lower():
                scheme, _, api_key = authorization.partition(" ")
            else:
                api_key = authorization
            if api_key != encryption_key:
                raise HTTPException(status_code=401, detail="Invalid API Key")
            return "USER"
        except Exception as e:
            raise HTTPException(status_code=401, detail="Invalid API Key")
    else:
        return "USER"


@app.get(
    "/v1/models",
    tags=["Models"],
    dependencies=[Depends(verify_api_key)],
)
async def models(user=Depends(verify_api_key)):
    return pipe.llm.models()


# For the completions and chat completions endpoints, we use extra_json for additional parameters.
# --------------------------------
# If `audio_format`` is present, the prompt will be transcribed to text.
#   It is assumed it is base64 encoded audio in the `audio_format`` specified.
# --------------------------------
# If `system_message`` is present, it will be used as the system message for the completion.
# --------------------------------
# If `voice`` is present, the completion will be converted to audio using the specified voice.
#   If not streaming, the audio will be returned in the response in the "audio" beside the "text" or "content" keys.
#   If streaming, the audio will be streamed in the response in audio/wav format.


# Chat completions endpoint
# https://platform.openai.com/docs/api-reference/chat
class ChatCompletions(BaseModel):
    model: str = DEFAULT_MODEL
    messages: List[dict] = None
    temperature: Optional[float] = 0.9
    top_p: Optional[float] = 1.0
    functions: Optional[List[dict]] = None
    function_call: Optional[str] = None
    n: Optional[int] = 1
    stream: Optional[bool] = False
    stop: Optional[List[str]] = None
    max_tokens: Optional[int] = 8192
    presence_penalty: Optional[float] = 0.0
    frequency_penalty: Optional[float] = 0.0
    logit_bias: Optional[Dict[str, float]] = None
    user: Optional[str] = None


class ChatCompletionsResponse(BaseModel):
    id: str
    object: str
    created: int
    model: str
    choices: List[dict]
    usage: dict


@app.post(
    "/v1/chat/completions",
    tags=["Completions"],
    dependencies=[Depends(verify_api_key)],
)
async def chat_completions(
    c: ChatCompletions, request: Request, user=Depends(verify_api_key)
):
    data = await request.json()
    response, audio_response = await pipe.get_response(
        data=data, completion_type="chat"
    )
    if audio_response:
        if audio_response.startswith("http"):
            return response
    if not c.stream:
        return response
    else:
        if audio_response:
            return StreamingResponse(
                content=audio_response,
                media_type="audio/wav",
            )
        return StreamingResponse(
            content=response["messages"][1]["content"],
            media_type="text/event-stream",
        )


# Completions endpoint
# https://platform.openai.com/docs/api-reference/completions
class Completions(BaseModel):
    model: str = DEFAULT_MODEL
    prompt: str = ""
    max_tokens: Optional[int] = 8192
    temperature: Optional[float] = 0.9
    top_p: Optional[float] = 1.0
    n: Optional[int] = 1
    stream: Optional[bool] = False
    logit_bias: Optional[Dict[str, float]] = None
    stop: Optional[List[str]] = None
    echo: Optional[bool] = False
    user: Optional[str] = None
    format_prompt: Optional[bool] = True


class CompletionsResponse(BaseModel):
    id: str
    object: str
    created: int
    model: str
    choices: List[dict]
    usage: dict


@app.post(
    "/v1/completions",
    tags=["Completions"],
    dependencies=[Depends(verify_api_key)],
)
async def completions(c: Completions, request: Request, user=Depends(verify_api_key)):
    response, audio_response = await pipe.get_response(
        data=await request.json(), completion_type="completion"
    )
    if audio_response:
        if audio_response.startswith("http"):
            return response
    if not c.stream:
        return response
    else:
        if audio_response:
            return StreamingResponse(
                content=audio_response,
                media_type="audio/wav",
            )
        return StreamingResponse(
            content=response["choices"][0]["text"],
            media_type="text/event-stream",
        )


# Embeddings endpoint
# https://platform.openai.com/docs/api-reference/embeddings
class EmbeddingModel(BaseModel):
    input: Union[str, List[str]]
    model: Optional[str] = DEFAULT_MODEL
    user: Optional[str] = None


class EmbeddingResponse(BaseModel):
    object: str
    data: List[dict]
    model: str
    usage: dict


@app.post(
    "/v1/embeddings",
    tags=["Embeddings"],
    dependencies=[Depends(verify_api_key)],
)
async def embedding(embedding: EmbeddingModel, user=Depends(verify_api_key)):
    return pipe.embedder.get_embeddings(input=embedding.input)


# Audio Transcription endpoint
# https://platform.openai.com/docs/api-reference/audio/createTranscription
@app.post(
    "/v1/audio/transcriptions",
    tags=["Audio"],
    dependencies=[Depends(verify_api_key)],
)
async def speech_to_text(
    file: UploadFile = File(...),
    model: str = Form(WHISPER_MODEL),
    language: Optional[str] = Form(None),
    prompt: Optional[str] = Form(None),
    response_format: Optional[str] = Form("json"),
    temperature: Optional[float] = Form(0.0),
    timestamp_granularities: Optional[List[str]] = Form(["segment"]),
    user: str = Depends(verify_api_key),
):
    response = await pipe.stt.transcribe_audio(
        base64_audio=base64.b64encode(await file.read()).decode("utf-8"),
        audio_format=file.content_type,
        language=language,
        prompt=prompt,
        temperature=temperature,
    )
    return {"text": response}


# Audio Translation endpoint
# https://platform.openai.com/docs/api-reference/audio/createTranslation


@app.post(
    "/v1/audio/translations",
    tags=["Audio"],
    dependencies=[Depends(verify_api_key)],
)
async def audio_translations(
    file: UploadFile = File(...),
    model: str = Form(WHISPER_MODEL),
    language: Optional[str] = Form(None),
    prompt: Optional[str] = Form(None),
    response_format: Optional[str] = Form("json"),
    temperature: Optional[float] = Form(0.0),
    user=Depends(verify_api_key),
):
    response = await pipe.stt.transcribe_audio(
        base64_audio=base64.b64encode(await file.read()).decode("utf-8"),
        audio_format=file.content_type,
        language=language,
        prompt=prompt,
        temperature=temperature,
        translate=True,
    )
    return {"text": response}


class TextToSpeech(BaseModel):
    input: str
    model: Optional[str] = "tts-1"
    voice: Optional[str] = "default"
    language: Optional[str] = "en"
    user: Optional[str] = None


@app.post(
    "/v1/audio/speech",
    tags=["Audio"],
    dependencies=[Depends(verify_api_key)],
)
async def text_to_speech(tts: TextToSpeech, user=Depends(verify_api_key)):
    if tts.input.startswith("data:"):
        if "pdf" in tts.input:
            audio = await pipe.pdf_to_audio(
                title=tts.user if tts.user else f"{uuid.uuid4().hex}",
                voice=tts.voice,
                pdf=tts.input,
                chunk_size=200,
            )
            return audio
        if "audio/" in tts.input:
            audio = await pipe.audio_to_audio(
                voice=tts.voice,
                audio=tts.input,
            )
            return audio
    audio = await pipe.ctts.generate(
        text=tts.input, voice=tts.voice, language=tts.language
    )
    return audio


@app.get(
    "/v1/audio/voices",
    tags=["Audio"],
    dependencies=[Depends(verify_api_key)],
)
async def get_voices(user=Depends(verify_api_key)):
    return {"voices": pipe.ctts.voices}


@app.post(
    "/v1/audio/voices",
    tags=["Audio"],
    dependencies=[Depends(verify_api_key)],
)
async def upload_voice(
    voice: str = Form("default"),
    file: UploadFile = File(...),
    user=Depends(verify_api_key),
):
    voice_name = voice
    file_path = os.path.join(os.getcwd(), "voices", f"{voice}.wav")
    if os.path.exists(file_path):
        i = 1
        while os.path.exists(file_path):
            file_path = os.path.join(os.getcwd(), "voices", f"{voice}-{i}.wav")
            voice_name = f"{voice}-{i}"
            i += 1
    with open(file_path, "wb") as audio_file:
        audio_file.write(await file.read())
    return {"detail": f"Voice {voice_name} has been uploaded."}


# Image Generation endpoint
# https://platform.openai.com/docs/api-reference/images


class ImageCreation(BaseModel):
    prompt: str
    model: Optional[str] = "stabilityai/sdxl-turbo"
    n: Optional[int] = 1
    size: Optional[str] = "512x512"
    quality: Optional[str] = "hd"
    response_format: Optional[str] = "url"
    style: Optional[str] = "natural"


@app.post(
    "/v1/images/generations",
    tags=["Images"],
    dependencies=[Depends(verify_api_key)],
)
async def generate_image(
    image_creation: ImageCreation,
    user: str = Depends(verify_api_key),
):
    images = []
    if int(image_creation.n) > 1:
        for i in range(image_creation.n):
            image = await pipe.generate_image(
                prompt=image_creation.prompt,
                response_format=image_creation.response_format,
                size=image_creation.size,
            )
            if image_creation.response_format == "url":
                images.append({"url": image})
            else:
                images.append({"b64_json": image})
        return {
            "created": int(time.time()),
            "data": images,
        }
    image = await pipe.generate_image(
        prompt=image_creation.prompt,
        response_format=image_creation.response_format,
        size=image_creation.size,
    )
    if image_creation.response_format == "url":
        return {
            "created": int(time.time()),
            "data": [{"url": image}],
        }
    return {
        "created": int(time.time()),
        "data": [{"b64_json": image}],
    }



import os
import base64
import uuid
import logging
import webrtcvad
import pyaudio
import wave
import threading
import numpy as np
from io import BytesIO
from faster_whisper import WhisperModel


class STT:
    def __init__(self, model="base", wake_functions={}):
        self.w = WhisperModel(model, download_root="models", device="cpu")
        self.audio = pyaudio.PyAudio()
        self.wake_functions = wake_functions

    async def transcribe_audio(
        self,
        base64_audio,
        audio_format="wav",
        language=None,
        prompt=None,
        temperature=0.0,
        translate=False,
    ):
        if "/" in audio_format:
            audio_format = audio_format.split("/")[1]
        filename = f"{uuid.uuid4().hex}.wav"
        file_path = os.path.join(os.getcwd(), "outputs", filename)
        audio_data = base64.b64decode(base64_audio)
        with open(file_path, "wb") as audio_file:
            audio_file.write(audio_data)
        if not os.path.exists(file_path):
            raise RuntimeError(f"Failed to load audio.")
        segments, _ = self.w.transcribe(
            file_path,
            task="transcribe" if not translate else "translate",
            vad_filter=True,
            vad_parameters=dict(min_silence_duration_ms=500),
            initial_prompt=prompt,
            language=language,
            temperature=temperature,
        )
        segments = list(segments)
        user_input = ""
        for segment in segments:
            user_input += segment.text
        logging.info(f"[STT] Transcribed User Input: {user_input}")
        os.remove(file_path)
        return user_input

    def listen(self):
        print("Listening for wake word...")
        vad = webrtcvad.Vad(1)
        stream = self.audio.open(
            format=pyaudio.paInt16,
            channels=1,
            rate=16000,
            input=True,
            frames_per_buffer=320,
        )
        frames = []
        silence_frames = 0
        while True:
            data = stream.read(320)
            frames.append(data)
            is_speech = vad.is_speech(data, 16000)
            if not is_speech:
                silence_frames += 1
                if silence_frames > 1 * 16000 / 320:
                    audio_data = b"".join(frames)
                    audio_np = np.frombuffer(audio_data, dtype=np.int16)
                    rms = np.sqrt(np.mean(audio_np**2))
                    if rms > 500:
                        buffer = BytesIO()
                        with wave.open(buffer, "wb") as wf:
                            wf.setnchannels(1)
                            wf.setsampwidth(self.audio.get_sample_size(pyaudio.paInt16))
                            wf.setframerate(16000)
                            wf.writeframes(b"".join(frames))
                        wav_buffer = buffer.getvalue()
                        base64_audio = base64.b64encode(wav_buffer).decode()
                        thread = threading.Thread(
                            target=self.transcribe_audio,
                            args=(base64_audio),
                        )
                        thread.start()
                    frames = []  # Clear frames after processing
                    silence_frames = 0
            else:
                silence_frames = 0



from llama_cpp import Llama, llama_chat_format
from huggingface_hub import hf_hub_download
from bs4 import BeautifulSoup
from typing import List, Optional, Dict
import os
import re
import requests
import psutil
import torch
import logging


DEFAULT_MODEL = os.environ.get("DEFAULT_MODEL", "TheBloke/phi-2-dpo-GGUF")


def get_vision_models():
    return [
        "mys/ggml_bakllava-1",
        "mys/ggml_llava-v1.5-7b",
        "mys/ggml_llava-v1.5-13b",
    ]


def get_models():
    response = requests.get("https://huggingface.co/models?sort=modified&search=gguf")
    soup = BeautifulSoup(response.text, "html.parser")
    model_names = []
    if soup:
        for a_tag in soup.find_all("a", href=True):
            href = a_tag["href"]
            if href.endswith("-GGUF"):
                model_names.append(href[1:])
    model_names.append(get_vision_models())
    return model_names


def is_vision_model(model_name="") -> bool:
    if model_name == "":
        return False
    model_name = model_name.lower()
    for model in get_vision_models():
        for key in model:
            if model_name == key.lower():
                return True
    return False


def download_llm(model_name="", models_dir="models"):
    global DEFAULT_MODEL
    model_name = model_name if model_name else DEFAULT_MODEL
    if "/" not in model_name:
        model_name = "TheBloke/" + model_name + "-GGUF"
    ram = round(psutil.virtual_memory().total / 1024**3)
    quantization_type = os.environ.get("QUANT_TYPE", "Q5_K_M" if ram > 16 else "Q4_K_M")
    model = model_name.split("/")[-1].split("-GGUF")[0]
    models_dir = os.path.join(models_dir, model)
    os.makedirs(models_dir, exist_ok=True)
    potential_clip_files = [
        "mmproj-model-f16.gguf",
        f"{model}-mmproj-f16.gguf",
    ]
    for clip_file in potential_clip_files:
        if not os.path.exists(f"{models_dir}/{clip_file}"):
            try:
                hf_hub_download(
                    repo_id=model_name,
                    filename=clip_file,
                    local_dir=models_dir,
                )
            except Exception as e:
                pass
    potential_filenames = [
        f"{model}.{quantization_type}.gguf",
        f"{model}-{quantization_type}.gguf",
        f"{model}.{quantization_type.lower()}.gguf",
        f"{model}-{quantization_type.lower()}.gguf",
        f"{model}.Q4_K_M.gguf",
        f"{model}.q4_k_m.gguf",
        f"{model}-Q4_K_M.gguf",
        f"{model}-q4_k_m.gguf",
        "ggml-model-q5_k.gguf",
        "ggml-model-f16.gguf",
    ]
    for filename in potential_filenames:
        filepath = os.path.join(models_dir, filename)
        if os.path.exists(filepath):
            return filepath
    logging.info(f"[LLM] Downloading {model}...")
    for filename in potential_filenames:
        filepath = os.path.join(models_dir, filename)
        try:
            hf_hub_download(
                repo_id=model_name,
                filename=filename,
                local_dir=models_dir,
            )
            logging.info(f"[LLM] Downloaded {model} successfully!")
            return filepath
        except Exception as e:
            pass
    raise FileNotFoundError("No suitable quantization file found.")


def get_clip_path(model_name="", models_dir="models"):
    if model_name == "":
        global DEFAULT_MODEL
        model_name = DEFAULT_MODEL
    potential_clip_files = [
        "mmproj-model-f16.gguf",
        f"{model_name.split('/')[-1]}-mmproj-f16.gguf",
    ]
    for clip_file in potential_clip_files:
        if os.path.exists(f"{models_dir}/{model_name}/{clip_file}"):
            return f"{models_dir}/{model_name}/{clip_file}"
    else:
        return ""


def clean(
    message: str = "",
    stop_tokens: List[str] = [
        "<|im_end|",
        "<|im_end|>",
        "</|im_end|>",
        "</s>",
        "<s>",
        "User:",
        "### \n###",
        "[/INST]",
        "<|eot_id|>",
        "<|end_of_text|>",
        "assistant\n\n",
    ],
):
    if message == "":
        return message
    for token in stop_tokens:
        if token in message:
            message = message.split(token)[0]
    message = message.strip()
    if message.startswith("\n "):
        message = message[3:]
    if message.endswith("\n\n"):
        message = message[:-4]
    if message.startswith(" "):
        message = message[1:]
    if message.endswith("\n"):
        message = message[:-3]
    if "[Insert " in message:
        message = re.sub(r"\[Insert.*?\]", "", message)
    return message


class LLM:
    def __init__(
        self,
        stop: List[str] = [],
        temperature: float = 1.31,
        max_tokens: int = 0,
        top_p: float = 0.95,
        min_p: float = 0.05,
        stream: bool = False,
        presence_penalty: float = 0.0,
        frequency_penalty: float = 0.0,
        logit_bias: Optional[Dict[str, float]] = None,
        model: str = "",
        models_dir: str = "./models",
        system_message: str = "",
        **kwargs,
    ):
        global DEFAULT_MODEL
        MAIN_GPU = os.environ.get("MAIN_GPU", 0)
        GPU_LAYERS = os.environ.get("GPU_LAYERS", 0)
        if torch.cuda.is_available() and int(GPU_LAYERS) == -1:
            # 5GB VRAM reserved for TTS and STT.
            vram = round(torch.cuda.get_device_properties(0).total_memory / 1024**3) - 5
            if vram == 3:
                vram = 1
            if vram <= 0:
                vram = 0
            logging.info(f"[LLM] {vram}GB of available VRAM detected.")
            GPU_LAYERS = vram - 1 if vram > 0 else 0
        if GPU_LAYERS == -2:
            GPU_LAYERS = -1
        logging.info(
            f"[LLM] Loading {DEFAULT_MODEL} with {GPU_LAYERS if GPU_LAYERS != -1 else 'all'} GPU layers. Please wait..."
        )
        self.params = {}
        self.model_name = DEFAULT_MODEL
        chat_handler = None
        if self.model_name != "":
            self.params["model_path"] = download_llm(
                model_name=self.model_name, models_dir=models_dir
            )
            if max_tokens != 0:
                self.params["max_tokens"] = 4096
            else:
                self.params["max_tokens"] = max_tokens
            if is_vision_model(model_name=self.model_name):
                clip_path = get_clip_path(
                    model_name=self.model_name, models_dir=models_dir
                )
                if clip_path != "":
                    chat_handler = llama_chat_format.Llava15ChatHandler(
                        clip_model_path=clip_path, verbose=True
                    )
        else:
            self.params["model_path"] = ""
            self.params["max_tokens"] = 8192
        self.params["n_ctx"] = int(os.environ.get("LLM_MAX_TOKENS", 4096))
        self.params["verbose"] = True
        self.system_message = system_message
        self.params["mirostat_mode"] = 2
        self.params["top_k"] = 20 if "top_k" not in kwargs else kwargs["top_k"]
        self.params["stop"] = [
            "<|im_end|",
            "<|im_end|>",
            "</|im_end|>",
            "</s>",
            "<s>",
            "User:",
            "### \n###",
            "[/INST]",
            "<|eot_id|>",
            "<|end_of_text|>",
            "assistant\n\n",
        ]
        if stop != []:
            if isinstance(stop, str):
                self.params["stop"].append(stop)
            else:
                try:
                    for stop_string in stop:
                        if stop_string not in self.params["stop"]:
                            if stop_string != None:
                                self.params["stop"].append(stop_string)
                except:
                    if stop != None:
                        self.params["stop"].append(stop)

        self.params["temperature"] = temperature if temperature else 1.31
        self.params["top_p"] = top_p if top_p else 0.95
        self.params["min_p"] = min_p if min_p else 0.05
        self.params["stream"] = stream if stream else False
        self.params["presence_penalty"] = presence_penalty if presence_penalty else 0.0
        self.params["frequency_penalty"] = (
            frequency_penalty if frequency_penalty else 0.0
        )
        self.params["repetition_penalty"] = 1.2
        self.params["logit_bias"] = logit_bias if logit_bias else None
        self.params["n_gpu_layers"] = int(GPU_LAYERS) if GPU_LAYERS else 0
        self.params["main_gpu"] = int(MAIN_GPU) if MAIN_GPU else 0
        if "batch_size" in kwargs:
            self.params["n_batch"] = (
                int(kwargs["batch_size"]) if kwargs["batch_size"] else 1024
            )
        else:
            self.params["n_batch"] = int(os.environ.get("LLM_BATCH_SIZE", 1024))
        if self.model_name != "":
            self.lcpp = Llama(
                **self.params,
                chat_handler=chat_handler,
                logits_all=True if chat_handler else False,
            )
        else:
            self.lcpp = None
        self.model_list = get_models()

    def generate(
        self,
        prompt,
        max_tokens=None,
        temperature=None,
        top_p=None,
        min_p=None,
        top_k=None,
        logit_bias=None,
        mirostat_mode=None,
        frequency_penalty=None,
        presence_penalty=None,
        stream=None,
        model=None,
        system_message=None,
        **kwargs,
    ):
        messages = [
            {
                "role": "system",
                "content": (
                    self.system_message if system_message is None else system_message
                ),
            }
        ]
        if isinstance(prompt, list):
            messages.append(
                {
                    "role": "user",
                    "content": prompt,
                },
            )
        else:
            messages.append(
                {
                    "role": "user",
                    "content": prompt,
                },
            )
        data = self.lcpp.create_chat_completion(
            messages=messages,
            max_tokens=(
                self.params["max_tokens"] if max_tokens is None else int(max_tokens)
            ),
            temperature=(
                self.params["temperature"]
                if temperature is None
                else float(temperature)
            ),
            top_p=self.params["top_p"] if top_p is None else float(top_p),
            min_p=self.params["min_p"] if min_p is None else float(min_p),
            stop=self.params["stop"],
            top_k=self.params["top_k"] if top_k is None else int(top_k),
            logit_bias=self.params["logit_bias"] if logit_bias is None else logit_bias,
            mirostat_mode=(
                self.params["mirostat_mode"]
                if mirostat_mode is None
                else int(mirostat_mode)
            ),
            frequency_penalty=(
                self.params["frequency_penalty"]
                if frequency_penalty is None
                else float(frequency_penalty)
            ),
            presence_penalty=(
                self.params["presence_penalty"]
                if presence_penalty is None
                else float(presence_penalty)
            ),
            stream=self.params["stream"] if stream is None else stream,
            model=self.model_name if model is None else model,
        )
        data["model"] = self.model_name
        return data

    def completion(self, prompt, **kwargs):
        data = self.generate(prompt=prompt, **kwargs)
        data["choices"][0]["message"]["content"] = clean(
            message=data["choices"][0]["message"]["content"],
            stop_tokens=self.params["stop"],
        )
        data["choices"][0]["text"] = data["choices"][0]["message"]["content"]
        return data

    def chat(self, messages, **kwargs):
        user_input = messages[-1]["content"]
        data = self.generate(prompt=user_input, **kwargs)
        data["choices"][0]["message"]["content"] = clean(
            message=data["choices"][0]["message"]["content"],
            stop_tokens=self.params["stop"],
        )
        return data

    def models(self):
        return self.model_list


if __name__ == "__main__":
    logging.info(f"[LLM] Downloading {DEFAULT_MODEL} model...")
    download_llm(model_name=DEFAULT_MODEL, models_dir="models")



import os
import logging
from dotenv import load_dotenv
from ezlocalai.LLM import LLM, is_vision_model
from ezlocalai.STT import STT
from ezlocalai.CTTS import CTTS
from ezlocalai.Embedding import Embedding
from pyngrok import ngrok
import requests
import base64
import pdfplumber
from typing import List

try:
    from ezlocalai.IMG import IMG

    img_import_success = True
except ImportError:
    img_import_success = False

from ezlocalai.VLM import VLM


class Pipes:
    def __init__(self):
        load_dotenv()
        global img_import_success
        DEFAULT_MODEL = os.getenv("DEFAULT_MODEL", "TheBloke/phi-2-dpo-GGUF")
        self.current_llm = DEFAULT_MODEL if DEFAULT_MODEL else "TheBloke/phi-2-dpo-GGUF"
        logging.info(f"[LLM] {self.current_llm} model loading. Please wait...")
        self.llm = LLM(model=self.current_llm)
        logging.info(f"[LLM] {self.current_llm} model loaded successfully.")
        self.embedder = Embedding()
        self.current_vlm = os.getenv("VISION_MODEL", "")
        logging.info(f"[VLM] {self.current_vlm} model loading. Please wait...")
        self.vlm = None
        if self.current_vlm != "":
            try:
                self.vlm = VLM(model=self.current_vlm)
            except Exception as e:
                logging.error(f"[VLM] Failed to load the model: {e}")
                self.vlm = None
        if self.vlm is not None:
            logging.info(f"[ezlocalai] Vision is enabled with {self.current_vlm}.")
        logging.info(f"[CTTS] xttsv2_2.0.2 model loading. Please wait...")
        self.ctts = CTTS()
        logging.info(f"[CTTS] xttsv2_2.0.2 model loaded successfully.")
        self.current_stt = os.getenv("WHISPER_MODEL", "base")
        logging.info(f"[STT] {self.current_stt} model loading. Please wait...")
        self.stt = STT(model=self.current_stt)
        logging.info(f"[STT] {self.current_stt} model loaded successfully.")
        if is_vision_model(self.current_llm):
            if self.vlm is None:
                self.vlm = self.llm
        NGROK_TOKEN = os.environ.get("NGROK_TOKEN", "")
        if NGROK_TOKEN:
            ngrok.set_auth_token(NGROK_TOKEN)
            public_url = ngrok.connect(8091)
            logging.info(f"[ngrok] Public Tunnel: {public_url.public_url}")
            self.local_uri = public_url.public_url
        else:
            self.local_uri = os.environ.get("EZLOCALAI_URL", "http://localhost:8091")
        self.img_enabled = os.getenv("IMG_ENABLED", "false").lower() == "true"
        self.img = None
        if img_import_success:
            logging.info(f"[IMG] Image generation is enabled.")
            SD_MODEL = os.getenv("SD_MODEL", "")  # stabilityai/sdxl-turbo
            if SD_MODEL:
                logging.info(f"[IMG] {SD_MODEL} model loading. Please wait...")
                try:
                    self.img = IMG(model=SD_MODEL, local_uri=self.local_uri)
                except Exception as e:
                    logging.error(f"[IMG] Failed to load the model: {e}")
                    self.img = None
            logging.info(f"[IMG] {SD_MODEL} model loaded successfully.")

    async def pdf_to_audio(self, title, voice, pdf, chunk_size=200):
        filename = f"{title}.pdf"
        file_path = os.path.join(os.getcwd(), "outputs", filename)
        pdf = pdf.split(",")[1]
        pdf = base64.b64decode(pdf)
        with open(file_path, "wb") as pdf_file:
            pdf_file.write(pdf)
        content = ""
        if file_path.endswith(".pdf"):
            with pdfplumber.open(file_path) as pdf:
                content = "\n".join([page.extract_text() for page in pdf.pages])
        if not content:
            return
        return await self.ctts.generate(
            text=content,
            voice=voice,
            local_uri=self.local_uri,
            output_file_name=f"{title}.wav",
        )

    async def audio_to_audio(self, voice, audio):
        audio_type = audio.split(",")[0].split(":")[1].split(";")[0]
        audio_format = audio_type.split("/")[1]
        audio = audio.split(",")[1]
        audio = base64.b64decode(audio)
        text = self.stt.transcribe_audio(base64_audio=audio, audio_format=audio_format)
        return await self.ctts.generate(
            text=text, voice=voice, local_uri=self.local_uri
        )

    async def generate_image(self, prompt, response_format="url", size="512x512"):
        if self.img:
            self.img.local_uri = self.local_uri if response_format == "url" else None
            new_image = self.img.generate(
                prompt=prompt,
                size=size,
            )
            self.img.local_uri = self.local_uri
            return new_image
        return ""

    async def get_response(self, data, completion_type="chat"):
        data["local_uri"] = self.local_uri
        images = []
        if "messages" in data:
            if isinstance(data["messages"][-1]["content"], list):
                messages = data["messages"][-1]["content"]
                for message in messages:
                    if "text" in message:
                        prompt = message["text"]
                for message in messages:
                    if "image_url" in message:
                        images.append(message)
                    if "audio_url" in message:
                        audio_url = (
                            message["audio_url"]["url"]
                            if "url" in message["audio_url"]
                            else message["audio_url"]
                        )
                        audio_format = "wav"
                        if audio_url.startswith("data:"):
                            audio_url = audio_url.split(",")[1]
                            audio_format = audio_url.split(";")[0]
                        else:
                            audio_url = requests.get(audio_url).content
                            audio_url = base64.b64encode(audio_url).decode("utf-8")
                        transcribed_audio = self.stt.transcribe_audio(
                            base64_audio=audio_url, audio_format=audio_format
                        )
                        prompt = f"Transcribed Audio: {transcribed_audio}\n\n{prompt}"
        if data["model"]:
            if self.current_llm != data["model"]:
                data["model"] = self.current_llm
        if "stop" in data:
            new_stop = self.llm.params["stop"]
            new_stop.append(data["stop"])
            data["stop"] = new_stop
        if "audio_format" in data:
            base64_audio = (
                data["messages"][-1]["content"]
                if completion_type == "chat"
                else data["prompt"]
            )
            prompt = await self.stt.transcribe_audio(
                base64_audio=base64_audio,
                audio_format=data["audio_format"],
            )
            if completion_type == "chat":
                data["messages"][-1]["content"] = prompt
            else:
                data["prompt"] = prompt
        user_message = (
            data["messages"][-1]["content"]
            if completion_type == "chat"
            else data["prompt"]
        )
        if self.vlm and images:
            new_messages = [
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "text",
                            "text": "Describe each stage of this image.",
                        },
                    ],
                }
            ]
            new_messages[0]["content"].extend(images)
            image_description = self.vlm.chat(messages=new_messages)
            print(
                f"Image Description: {image_description['choices'][0]['message']['content']}"
            )
            prompt = (
                f"\n\nSee the uploaded image description for any questions about the uploaded image. Act as if you can see the image based on the description. Do not mention 'uploaded image description' in response. Uploaded Image Description: {image_description['choices'][0]['message']['content']}\n\n{data['messages'][-1]['content'][0]['text']}"
                if completion_type == "chat"
                else f"\n\nSee the uploaded image description for any questions about the uploaded image. Act as if you can see the image based on the description. Do not mention 'uploaded image description' in response. Uploaded Image Description: {image_description['choices'][0]['message']['content']}\n\n{data['prompt']}"
            )
            print(f"Full Prompt: {prompt}")
            if completion_type == "chat":
                data["messages"][-1]["content"] = prompt
            else:
                data["prompt"] = prompt
        if completion_type == "chat":
            response = self.llm.chat(**data)
        else:
            response = self.llm.completion(**data)
        generated_image = None
        if "temperature" not in data:
            data["temperature"] = 0.5
        if "top_p" not in data:
            data["top_p"] = 0.9
        if self.img_enabled and img_import_success and self.img:
            user_message = (
                data["messages"][-1]["content"]
                if completion_type == "chat"
                else data["prompt"]
            )
            if isinstance(user_message, list):
                user_message = prompt
                for message in messages:
                    if "image_url" in message:
                        if "url" in message["image_url"]:
                            if not message["image_url"]["url"].startswith("data:"):
                                user_message += (
                                    "Uploaded Image:"
                                    + message["image_url"]["url"]
                                    + "\n"
                                )
            response_text = (
                response["choices"][0]["text"]
                if completion_type != "chat"
                else response["choices"][0]["message"]["content"]
            )
            if "data:" in user_message:
                user_message = user_message.replace(
                    user_message.split("data:")[1].split("'")[0], ""
                )
            img_gen_prompt = f"Users message: {user_message} \n\n{'The user uploaded an image, one does not need generated unless the user is specifically asking.' if images else ''} **The assistant is acting as sentiment analysis expert and only responds with a concise YES or NO answer on if the user would like an image as visual or a picture generated. No other explanation is needed!**\nWould the user potentially like an image generated based on their message?\nAssistant: "
            logging.info(f"[IMG] Decision maker prompt: {img_gen_prompt}")
            create_img = self.llm.chat(
                messages=[{"role": "system", "content": img_gen_prompt}],
                max_tokens=10,
                temperature=data["temperature"],
                top_p=data["top_p"],
            )
            create_img = str(create_img["choices"][0]["message"]["content"]).lower()
            logging.info(f"[IMG] Decision maker response: {create_img}")
            if "yes" in create_img or "es," in create_img:
                img_prompt = f"**The assistant is acting as a Stable Diffusion Prompt Generator.**\n\nUsers message: {user_message} \nAssistant response: {response_text} \n\nImportant rules to follow:\n- Describe subjects in detail, specify image type (e.g., digital illustration), art style (e.g., steampunk), and background. Include art inspirations (e.g., Art Station, specific artists). Detail lighting, camera (type, lens, view), and render (resolution, style). The weight of a keyword can be adjusted by using the syntax (((keyword))) , put only those keyword inside ((())) which is very important because it will have more impact so anything wrong will result in unwanted picture so be careful. Realistic prompts: exclude artist, specify lens. Separate with double lines. Max 60 words, avoiding 'real' for fantastical.\n- Based on the message from the user and response of the assistant, you will need to generate one detailed stable diffusion image generation prompt based on the context of the conversation to accompany the assistant response.\n- The prompt can only be up to 60 words long, so try to be concise while using enough descriptive words to make a proper prompt.\n- Following all rules will result in a $2000 tip that you can spend on anything!\n- Must be in markdown code block to be parsed out and only provide prompt in the code block, nothing else.\nStable Diffusion Prompt Generator: "
                image_generation_prompt = self.llm.chat(
                    messages=[{"role": "system", "content": img_prompt}],
                    max_tokens=100,
                    temperature=data["temperature"],
                    top_p=data["top_p"],
                )
                image_generation_prompt = str(
                    image_generation_prompt["choices"][0]["message"]["content"]
                )
                logging.info(
                    f"[IMG] Image generation response: {image_generation_prompt}"
                )
                if "```markdown" in image_generation_prompt:
                    image_generation_prompt = image_generation_prompt.split(
                        "```markdown"
                    )[1]
                    image_generation_prompt = image_generation_prompt.split("```")[0]
                generated_image = self.img.generate(prompt=image_generation_prompt)
        audio_response = None
        if "voice" in data:
            text_response = (
                response["choices"][0]["text"]
                if completion_type != "chat"
                else response["choices"][0]["message"]["content"]
            )
            language = data["language"] if "language" in data else "en"
            audio_response = await self.ctts.generate(
                text=text_response,
                voice=data["voice"],
                language=language,
                local_uri=self.local_uri,
            )
            if completion_type != "chat":
                response["choices"][0]["text"] = f"{text_response}\n{audio_response}"
            else:
                response["choices"][0]["message"][
                    "content"
                ] = f"{text_response}\n{audio_response}"
        if generated_image:
            if completion_type != "chat":
                response["choices"][0]["text"] += f"\n\n{generated_image}"
            else:
                response["choices"][0]["message"]["content"] += f"\n\n{generated_image}"
        return response, audio_response



import logging
import uuid
import torch
from PIL import Image

try:
    from diffusers import DiffusionPipeline, LCMScheduler

    import_success = True
except ImportError:
    logging.error(
        "Failed to import diffusers. Please install diffusers using 'pip install diffusers'"
    )
    import_success = False


class IMG:
    def __init__(
        self,
        model="stabilityai/sdxl-turbo",
        device="cpu",
        local_uri=None,
    ):
        global import_success
        self.local_uri = local_uri
        if import_success:
            pipe = DiffusionPipeline.from_pretrained(
                model,
                cache_dir="models",
                torch_dtype=torch.float32,
                scheduler=LCMScheduler(beta_start=0.001, beta_end=0.01),
            ).to(device)
            self.pipe = pipe
            self.pipe.enable_attention_slicing()
            self.pipe.safety_checker = None
        else:
            self.pipe = None

    def generate(
        self,
        prompt,
        negative_prompt="low resolution, grainy, distorted",
        num_inference_steps=1,
        guidance_scale=0.0,
        size="512x512",
    ):
        new_file_name = f"outputs/{uuid.uuid4()}.png"
        if self.pipe:
            generator = torch.Generator(device="cpu").manual_seed(0)
            new_image = self.pipe(
                prompt=prompt,
                negative_prompt=negative_prompt,
                num_inference_steps=num_inference_steps,
                guidance_scale=guidance_scale,
                generator=generator,
            ).images[0]
            width, height = map(int, size.split("x"))
            if width != 512 and height != 512:
                upscaled_image = new_image.resize(
                    (width, height), resample=Image.LANCZOS
                )
                upscaled_image.save(new_file_name)
            else:
                new_image.save(new_file_name)
            if self.local_uri:
                return f"{self.local_uri}/{new_file_name}"
            return upscaled_image
        else:
            return None
